# todo: move these to "data_dir" these constants shouldn't be included in any package
import numpy as np
mimic_thresholds = {
    'Diastolic blood pressure': [80, 90, 120],
    'Fraction inspired oxygen': [0.5],
    'Glascow coma scale total':[9, 13],
    'Glucose': [140, 200], 
    'Heart Rate': [60, 100, 140],
    'Mean blood pressure': [70, 100, 115, 150],
    'Oxygen saturation': [90, 95],
    'Respiratory rate': [12, 16, 20],
    'Systolic blood pressure': [120, 140, 180],
    'Temperature': [36.0, 37.0, 38.0], # deg C
    'Weight': [63, 83, 100], # kg
    'pH': [7.35, 7.45],
    'Age': [34, 49, 65, 79],
    'Gender': ['M', 'F']
}

def build_threshold_dict(col_names, mimic_thresholds):
    thresholds = {}
    for i in col_names:
        if i in mimic_thresholds:
            thresholds[i] = mimic_thresholds[i]
        elif i.endswith('count'):
            thresholds[i] = [1]
        elif i[2:-5] in mimic_thresholds:
            thresholds[i] = mimic_thresholds[i[2:-5]]
        elif i[:-5] in mimic_thresholds:
            thresholds[i] = mimic_thresholds[i[:-5]]
        else:
            print(i)
    return thresholds    

seq_grids = {
    'heart': {
        'N': list(range(1, 14)),
        'M': list(range(1, 6)),
        'FNR': np.arange(0.1, 1.01, 0.1)        
    }   
}
seq_grids['mort'] = seq_grids['heart']


# This file contains paths to key directories
from pathlib import Path

# path to the GitHub repository
repo_dir = Path(__file__).resolve().parent.parent

# path to the Python package
pkg_dir = repo_dir / 'IPChecklists'

# directory where we keep data
data_dir = repo_dir / 'data'



import pandas as pd
import numpy as np
from pandas.api.types import is_string_dtype
import random
from sklearn.metrics import accuracy_score, confusion_matrix, roc_auc_score, average_precision_score
from cplex.callbacks import IncumbentCallback, HeuristicCallback, MIPInfoCallback
from imblearn.over_sampling import RandomOverSampler
from sklearn.model_selection import StratifiedKFold

def infer_types(df, subset = None):
    '''
    input: df[subset], a dataframe of features
    output: a dictionary {col: type}, where type in ['categorical', 'ordinal', numeric']
    '''
    temp = df if subset is None else df[subset]
    res = {}
    for i in temp:
        if is_string_dtype(temp[i]) or len(temp[i].unique()) == 2:
            res[i] = 'categorical'
        elif temp[i].dtype in (np.int32, np.int64, int) and len(temp[i].unique()) < 50:
            res[i] = 'ordinal'
        else:
            res[i] = 'numeric'
    return res

def random_supersample(df, target_name):
    rs = RandomOverSampler(sampling_strategy='minority', random_state = 42)
    return df.loc[rs.fit_resample(df.index.values.reshape(-1, 1), df[target_name] == df[target_name].iloc[0])[0].squeeze()]

def capsCase(st):
    temp = st.split(' ')
    return ' '.join([i[0].upper() + i[1:] for i in temp])

def set_seed(seed = 42):
    random.seed(seed)
    np.random.seed(seed)

def set_seed_wrapper(f, seed = 42):
    def mod_f(*args, **kwargs):
        st0 = np.random.get_state()
        st1 = random.getstate()
        np.random.seed(seed)
        random.seed(seed)
        ans = f(*args, **kwargs)
        np.random.set_state(st0)
        random.setstate(st1)
        return ans

    return mod_f

def split_folds(df, target_name, fold, random_seed = 42):
    skf = StratifiedKFold(n_splits = 5, random_state = random_seed, shuffle = True)
    df['fold_id'] = None
    for c, (_, fold_idx) in enumerate(skf.split(list(range(df.shape[0])), df[target_name])):
        df.iloc[fold_idx, df.columns.get_loc('fold_id')] = c + 1

    if fold == 0:
        return (df.drop(columns = ['fold_id']), df.drop(columns = ['fold_id']))
    else:
        return (df[df.fold_id != fold].drop(columns = ['fold_id']), df[df.fold_id == fold].drop(columns = ['fold_id']))
    

def CIs(row):
    return {
        'mean': row.mean(),
        'lower': np.quantile(row, 0.025),
        'upper': np.quantile(row, 0.975)
    }

def bin_var(bins, x):
    for c, i in enumerate(bins):
        if (x >= i[0] and x < i[1]) or (c == len(bins) -1 and x == i[1]):
            return f'{i[0]}-{i[1]}'

def binary_metrics(targets, preds):    
    if len(targets) == 0:
        return {}
    res = {'accuracy': accuracy_score(targets, preds)}
    CM = confusion_matrix(targets, preds, labels = [0, 1])

    res['n_samples'] = len(targets)

    res['TN'] = np.asscalar(CM[0][0])
    res['FN'] = np.asscalar(CM[1][0])
    res['TP'] = np.asscalar(CM[1][1])
    res['FP'] = np.asscalar(CM[0][1])

    res['error'] = res['FN'] + res['FP']

    if res['TP']+res['FN'] == 0:
        res['TPR'] = 0
        res['FNR'] = 1
    else:
        res['TPR'] = res['TP']/(res['TP']+res['FN'])
        res['FNR'] = res['FN']/(res['TP']+res['FN'])

    if res['FP']+res['TN'] == 0:
        res['FPR'] = 1
        res['TNR'] = 0
    else:
        res['FPR'] = res['FP']/(res['FP']+res['TN'])
        res['TNR'] = res['TN']/(res['FP']+res['TN'])

    res['targets'] = targets
    res['preds'] = preds
    res['precision'] = res['TP'] / (res['TP'] + res['FP']) if (res['TP'] + res['FP']) > 0 else 0
    res['pred_prevalence'] = (res['TP'] + res['FP'])/res['n_samples']    
    res['prevalence'] = (res['TP'] + res['FN'])/res['n_samples']  
    
    return res

def probability_metrics(targets, preds):    
    return{
        'n_samples': len(targets),
        'AUROC': roc_auc_score(targets, preds),
        'AUPRC': average_precision_score(targets, preds)
    }    

def check_group_constraint(checklist, feature_mapping):
    selected_cols = set(checklist.get_selected_cols())
    for i in feature_mapping:
        if len(selected_cols.intersection(set(feature_mapping[i].binarized_names))) > 1:
            return False    
    return True            

def convert_target_domain(target):
    if np.all(np.sort(np.unique(target)) == np.array([-1, +1])):
        return (target + 1)/2
    else:
        return target

def per_group_metrics(ds, preds, attributes):
    assert(all([i in ds.protected_attrs for i in attributes]))
    rows = {}
    for attr in attributes:   
        for grp in ds.protected_groups[attr].unique():
            mask = ds.protected_groups[attr] == grp
            preds_i = preds[mask]
            targets_i = convert_target_domain(ds.target[mask])
            metrics_i = binary_metrics(targets_i, preds_i)
            rows[grp] = metrics_i
    return pd.DataFrame(rows)

def get_max_disparity(df, grps, metrics = ['FPR', 'FNR'], prefix = 'training'):
    '''
    Given a dataframe output from per_group_metrics, returns the max disparity and largest value across grps for metrics
    '''
    res = {}
    for met in metrics:
        row = df.loc[met, grps]
        res[prefix + '_' + met + ' (Maximum Disparity)'] = np.max(row) - np.min(row) # largest gap
        res[prefix + '_' + met + ' (Largest)'] = np.max(row) # worst value
    return pd.Series(res)

class StatsCallback(MIPInfoCallback):

    def initialize(self, store_solutions = False, solution_start_idx = None, solution_end_idx = None):

        # scalars
        self.times_called = 0
        self.start_time = None

        # stats that are stored at every call len(stat) = times_called
        self.runtimes = []
        self.nodes_processed = []
        self.nodes_remaining = []
        self.lowerbounds = []

        # stats that are stored at every incumbent update
        self.best_objval = float('inf')
        self.update_iterations = []
        self.incumbents = []
        self.upperbounds = []

        self.store_incumbent_solutions = store_solutions

        if self.store_incumbent_solutions:
            assert solution_start_idx <= solution_end_idx
            self.start_idx, self.end_idx = int(solution_start_idx), int(solution_end_idx)
            self.process_incumbent = self.record_objval_and_solution_before_incumbent
        else:
            self.process_incumbent = self.record_objval_before_incumbent

    def __call__(self):
        self.times_called += 1
        if self.start_time is None:
            self.start_time = self.get_start_time()
        self.runtimes.append(self.get_time())
        self.lowerbounds.append(self.get_best_objective_value())
        self.nodes_processed.append(self.get_num_nodes())
        self.nodes_remaining.append(self.get_num_remaining_nodes())
        self.process_incumbent()

    def record_objval_before_incumbent(self):
        if self.has_incumbent():
            self.record_objval()
            self.process_incumbent = self.record_objval

    def record_objval_and_solution_before_incumbent(self):
        if self.has_incumbent():
            self.record_objval_and_solution()
            self.process_incumbent = self.record_objval_and_solution

    def record_objval(self):
        objval = self.get_incumbent_objective_value()
        if objval < self.best_objval:
            self.best_objval = objval
            self.update_iterations.append(self.times_called)
            self.upperbounds.append(objval)

    def record_objval_and_solution(self):
        objval = self.get_incumbent_objective_value()
        if objval < self.best_objval:
            self.update_iterations.append(self.times_called)
            self.incumbents.append(self.get_incumbent_values(self.start_idx, self.end_idx))
            self.upperbounds.append(objval)
            self.best_objval = objval

    def check_stats(self):
        """checks stats rep at any point during the solution process"""

        # try:
        n_calls = len(self.runtimes)
        n_updates = len(self.upperbounds)
        assert n_updates <= n_calls

        if n_calls > 0:

            assert len(self.nodes_processed) == n_calls
            assert len(self.nodes_remaining) == n_calls
            assert len(self.lowerbounds) == n_calls

            lowerbounds = np.array(self.lowerbounds)
            for ub in self.upperbounds:
                assert np.all(ub + 1e-5 >= lowerbounds)

            runtimes = np.array(self.runtimes) - self.start_time
            nodes_processed = np.array(self.nodes_processed)

            is_increasing = lambda x: (np.diff(x) >= -1e-5).all()
            assert is_increasing(runtimes)
            assert is_increasing(nodes_processed)
            assert is_increasing(lowerbounds)

        if n_updates > 0:

            assert len(self.update_iterations) == n_updates
            if self.store_incumbent_solutions:
                assert len(self.incumbents) == n_updates

            update_iterations = np.array(self.update_iterations)
            upperbounds = np.array(self.upperbounds)
            gaps = (upperbounds - lowerbounds[update_iterations - 1]) / (np.finfo(np.float).eps + upperbounds)

            is_increasing = lambda x: (np.diff(x) >= -1e-5).all()
            assert is_increasing(update_iterations)
            assert is_increasing(-upperbounds)
            assert is_increasing(-gaps)

        # except AssertionError:
        #    ipsh()

        return True

    def get_stats(self, return_solutions = False):

        assert self.check_stats()
        import pandas as pd
        MAX_UPPERBOUND = float('inf')
        MAX_GAP = 1.00

        stats = pd.DataFrame({
            'runtime': [t - self.start_time for t in self.runtimes],
            'nodes_processed': list(self.nodes_processed),
            'nodes_remaining': list(self.nodes_remaining),
            'lowerbound': list(self.lowerbounds)
            })

        upperbounds = list(self.upperbounds)
        update_iterations = list(self.update_iterations)
        incumbents = [] #empty placeholder            
            
        if len(update_iterations) == 0:
            return stats, incumbents
            
        # add upper bounds as well as iterations where the incumbent changes

        if update_iterations[0] > 1:
            update_iterations.insert(0, 1)
            upperbounds.insert(0, MAX_UPPERBOUND)
        row_idx = [i - 1 for i in update_iterations]

        stats = stats.assign(iterations = pd.Series(data = update_iterations, index = row_idx),
                             upperbound = pd.Series(data = upperbounds, index = row_idx))
        stats['incumbent_update'] = np.where(~np.isnan(stats['iterations']), True, False)
        stats = stats.fillna(method = 'ffill')

        # add relative gap
        gap = (stats['upperbound'] - stats['lowerbound']) / (stats['upperbound'] + np.finfo(np.float).eps)
        stats['gap'] = np.fmin(MAX_GAP, gap)

        # add model ids
        if return_solutions:
            incumbents = list(self.incumbents)
            model_ids = range(len(incumbents))
            row_idx = [i - 1 for i in self.update_iterations]
            stats = stats.assign(model_ids = pd.Series(data = model_ids, index = row_idx))
            stats = stats[['runtime',
                           'gap',
                           'upperbound',
                           'lowerbound',
                           'nodes_processed',
                           'nodes_remaining',
                           'model_id',
                           'incumbent_update']]

        else:
            stats = stats[['runtime',
                           'gap',
                           'upperbound',
                           'lowerbound',
                           'nodes_processed',
                           'nodes_remaining']]

        return stats, incumbents

class ProcessedDataset():
    def __init__(self, mat, target, I_plus, I_minus, n):
        self.mat = mat
        self.target = target
        self.n = n
        self.I_plus = I_plus
        self.I_minus = I_minus
        self.n_samples_plus = self.n[self.I_plus].sum()
        self.n_samples_minus = self.n[self.I_minus].sum()
        self.n_samples = self.n_samples_plus + self.n_samples_minus
        self.n_features = mat.shape[1]



import numpy as np
import pandas as pd
import copy
from IPChecklists.Checklist import Checklist
import math
from tqdm import tqdm

def cardinality_with_matroid(S, K, f, matroid = None, *args, **kwargs):
    '''
    S (list): list of feature indices
    K (int): cardinality constraint
    f (function): submodular function to maximize, should be parallelized across samples
    matroid (function): function that takes in a set of indices 
        and returns whether matroid is satisfied
    other args: passed to f (ex: M)
    '''
    all_sgs, all_func_vals = [], []
    sg = []
    while len(sg) < K and len(S) > 0:
        func_vals = f(sg, S, *args, **kwargs)
        if matroid is not None:
            satisfies_p = np.array(list(map(lambda x: matroid(sg + [x]), S)))
        else:
            satisfies_p = np.array([True] * len(S))

        x_star_ind = np.arange(len(S))[np.argmax(func_vals)]
        if satisfies_p[x_star_ind]:
            sg.append(S[x_star_ind]) 
            if len(all_func_vals) > 0 and func_vals[x_star_ind] == all_func_vals[-1]: # already maximized function
                break

            all_sgs.append(copy.deepcopy(sg))
            all_func_vals.append(func_vals[x_star_ind]) 

        if isinstance(S, np.ndarray):        
            S = np.delete(S, x_star_ind)
        else:
            del S[x_star_ind]

    return (all_sgs, all_func_vals)      


def knapsack_with_p_system(S, K, f, c, budget, eps, p_system = None, f_parallel = None, *args, **kwargs):
    '''
    S (list): list of feature indices
    K (int): cardinality constraint
    f (function): submodular function to maximize    
    c (list): cost of each item for knapsack constraint
    budget (float): budget for knapsack constraint
    eps (float): step size for algorithm
    p_system (function): function that takes in a set of indices 
        and returns whether p system is satisfied
    f_parallel (function): f but parallelized over some batch size
    other args: passed to f and f_parallel (ex: M)
    '''
    assert((S == np.arange(len(S))).all())
    all_sgs, all_func_vals = [], []
    if f_parallel is not None:
        m = max(f_parallel([], S, *args, **kwargs))
    else:
        m = max(list(map(lambda x: f([x], *args, **kwargs), S)))
    p = 2 if p_system is not None else 1 # cardinality + partition matroid
    l = 1 # one knapsack 
    n = len(S)
    c = c/float(budget)  
                
    rho_seq = [m/(p+l)]
    while rho_seq[-1] < 2*m*n/(p+l):
        rho_seq.append((1+eps)*rho_seq[-1])    
                
    rho_results = {}    
    for rho in rho_seq:
        if f_parallel is not None:
            tau_list = f_parallel([], S, *args, **kwargs)
        else:
            tau_list = list(map(lambda x: f([x], *args, **kwargs), S))
        tau_list = [tau_list[j] for j in range(len(S)) if c[j] == 0 or tau_list[j]/c[j] >= rho]
        if len(tau_list) == 0:
            break
                    
        tau = max(tau_list)        
        sg = []
        tp, tpp = None, None
        
        restart_next_rho = False        
        
        while tau >= eps * m/n and compute_cost(c, sg) <= 1 and not restart_next_rho:
            if f_parallel is not None:
                fs = f_parallel(sg, S, *args, **kwargs) - f(sg, *args, **kwargs)
                
            for j_ind, j in enumerate(S):
                if f_parallel is None:
                    fs_j = f(sg + [j], *args, **kwargs) - f(sg, *args, **kwargs)
                else:
                    fs_j = fs[j_ind]
                
                if ((p_system is None or p_system(sg + [j])) 
                        and fs_j  >= tau 
                        and (c[j_ind] == 0 or fs_j / c[j_ind] >= rho)
                        and len(sg) + 1 <= K) :
                    sg.append(j)
                    if compute_cost(c, sg) > 1:
                        tp = sg[:-1]
                        tpp = [sg[-1]]
                        restart_next_rho = True                        
                        break
                    else:
                        # recompute fs since sg is now different
                        if f_parallel is not None:
                            fs = f_parallel(sg, S, *args, **kwargs) - f(sg, *args, **kwargs)                        
                        
            tau *= 1/(1+eps)            
        
        if not restart_next_rho:
            tp = sg
        
        if tp is not None and len(tp) > 0: 
            rho_results[rho] = [tp]
            if tpp is not None and len(tpp) > 0:
                rho_results[rho].append(tpp)
                
    return rho_results        
    
def compute_cost(c, sg):
    return np.take(c, sg).sum()

def query_function(S, M, X):
    # s is a list of indices of features
    weights = np.zeros((X.shape[1],))
    weights[S] = 1
    return np.sum(np.minimum(np.dot(X, weights), M))

def query_function_parallel(sg, S, M, X):
    # sg (list): list of features indices currently in the set
    # S (list): list of feature indices that can be added
    weights = np.zeros((X.shape[1], len(S)))
    weights[sg, :] = 1
    weights[S, np.arange(weights.shape[1])] = 1    
    return np.sum(np.minimum(np.dot(X, weights), [M] * len(S)), axis = 0)

def satisfy_constraints(s, splits):
    '''
    s (list): subset of features to check
    splits (list of sets): allow at most one feature from each set
    '''    
    s = set(s)
    for orig in splits:
        if len(orig.intersection(s)) > 1:
            return False
    return True   

def remove_duplicates(results):
    indices_sets, result_covers = [], []
    for cover_result in results:
        index_set = set(list(cover_result.indices))
        if index_set not in indices_sets and cover_result.M <= cover_result.N:
            indices_sets.append(index_set)
            result_covers.append(cover_result)
    return result_covers

class Cover():
    def __init__(self, ds, M, max_N = None, cover_label = 1, enforce_group_constraint = False):
        super().__init__()
        self.ds = ds        
        self.M = M
        self.max_N = len(ds.binarized_df.shape[1]) if max_N is None else max_N
        self.cover_label = cover_label
        self.enforce_group_constraint = enforce_group_constraint
    
    def get_results(self, return_intermediate = True):
        if return_intermediate:
            return self.results
        return self.results[-1]

    def get_X(self):
        if self.cover_label == 1:
            X = self.ds.binarized_df[self.ds.target == 1]      
        elif self.cover_label == -1:  
            X = -self.ds.binarized_df[self.ds.target == -1] + 1
        else:
            raise ValueError
        return X.values

    def get_matroid(self):
        if self.enforce_group_constraint:
            splits = []
            for feature_set in self.ds.feature_mapping.values():
                inds = set()
                for c, i in enumerate(self.ds.binarized_df.columns):
                    if i in feature_set.binarized_names:
                        inds.add(c)
                splits.append(inds)
            matroid = lambda s: satisfy_constraints(s, splits)
        else:
            matroid = None
        return matroid


class SimpleCover(Cover):
    # cardinality constraint with or without p-system constraints
    def __init__(self, ds, M, max_N = None, cover_label = 1, enforce_group_constraint = True):        
        super().__init__(ds, M, max_N, cover_label, enforce_group_constraint)
        
    def solve(self):
        X = self.get_X()
        matroid = self.get_matroid()
            
        sgs, vals = cardinality_with_matroid(np.arange(X.shape[1]), self.max_N, query_function_parallel,
                                    matroid = matroid, M = self.M, X = X)
        max_val = np.max(vals)
        upper_limit_orig = int(max_val / (1-1/math.e))
        # number of positive/negatives that can be covered in the original MIP
        upper_limit = max(0, upper_limit_orig - (self.M - 1) * (self.ds.target == self.cover_label).sum())            
        self.results = remove_duplicates([CoverResult(sg, self.M, val, upper_limit, self.ds) for sg, val in zip(sgs, vals)])          

class KnapsackCover(Cover):
    # knapsack constraint + cardinality constraint with/without p-system constraints
    def __init__(self, ds, M, budget, eps, max_N = None, cover_label = 1, enforce_group_constraint = True):
        super().__init__(ds, M, max_N, cover_label, enforce_group_constraint)
        self.budget = budget
        self.eps = eps
        self.p = int(enforce_group_constraint)
        
    def solve(self):
        X = self.get_X()                
        c = (self.ds.binarized_df.values[self.ds.target == -self.cover_label] == 1).sum(axis = 0).astype(float)
            
        ratio = self.p + 2*2 + 1 + self.eps # l = 2
        p_system = self.get_matroid()

        results = knapsack_with_p_system(np.arange(X.shape[1]), self.max_N, query_function,
                                        c = c, budget = self.budget, eps = self.eps, p_system = p_system,
                                                f_parallel = query_function_parallel, M = self.M, X = X)
        sgs, vals = [], []
        for rho in results:
            for lst in results[rho]:
                if len(lst) > 0:
                    sgs.append(lst)
                    vals.append(query_function(lst, self.M, X))
        max_val = np.max(vals)            
        upper_limit_orig = int(max_val * ratio)
        # number of positive/negatives that can be covered in the original MIP
        upper_limit = max(0, upper_limit_orig - (self.M - 1) * (self.ds.target == self.cover_label).sum()) 
        
        self.results = remove_duplicates([CoverResult(sg, self.M, val, upper_limit, self.ds) for sg, val in zip(sgs, vals)])

class CoverResult():
    def __init__(self, indices, M, func_val, upper_limit, ds):
        self.N = len(indices)
        self.indices = indices
        self.lamb = np.zeros((ds.binarized_df.shape[1],))
        self.lamb[indices] = 1
        self.func_val = func_val
        self.M = M
        self.upper_limit = upper_limit
        self.column_names = ds.binarized_df.columns
        
    def to_checklist(self):
        return Checklist(from_weights = True, lamb = self.lamb, M = self.M, column_names = self.column_names)

    def __eq__(self, other):
        return set(list(self.indices)) == set(list(other.indices))


import numpy as np
import pandas as pd
from .utils import infer_types, set_seed_wrapper
import logging
import optbinning
import copy

logging.basicConfig(level='INFO')
log = logging.getLogger()

class Feature():
    def __init__(self, overall_name, binarized_names, dtype, complement_pairs = []):
        self.overall_name = overall_name
        self.binarized_names = binarized_names
        self.type = dtype
        self.count = len(binarized_names)
        self.complement_pairs = complement_pairs

def compr(x, val, op):
    if op == '>=':
        return x >= val
    elif op == '>':
        return x > val
    elif op == '<=':
        return x <= val
    elif op == '<':
        return x < val
    elif op == '==':
        return x == val
    else:
        raise NotImplementedError(op)

inverse_ops = {
     '==': '!=',
     '>=': '<',
    '<=': '>',
    '!=': '=='
}

def compute_optbin_thresholds(df, dtypes, target, max_n_bins = 5):
    target = np.copy(target)
    target[target == -1] = 0
    thresholds = {}
    for col in dtypes:
        optb = set_seed_wrapper(optbinning.OptimalBinning)(name = col, dtype = dtypes[col] if dtypes[col] == 'categorical' else 'numerical',
                                         solver="cp", max_n_bins = max_n_bins if dtypes[col] == 'numeric' else None)
        optb.fit(df[col].values, target)
        thresholds[col] = list(optb.splits)

        if dtypes[col] == 'categorical':
            thresholds[col] = [j for i in thresholds[col] for j in i]

    return thresholds


class BinaryDataset():
    '''
    Given a dataframe of features and a single target, binarizes the features
        and prepares it for checklist training

    Potentially useful attributes:
       - binarized_df
       - target
       - col_types
       - thresholds
       - feature_mapping
       - remaining_features
    '''
    def __init__(self, df, target_name, pos_label, col_subset = None, percentile_inc = 25, op = '>=',
                    already_binarized = False, thresholds = None, add_complements = False, use_optbin = False,
                 max_n_bins = 5, complement_includes_missing = True, use_missing_features = True,
                 protected_attrs = [], categorical_bin = 'all'):
        '''
        Args:
            df (Dataframe): dataframe with features for modelling, and `target_name`
            target_name (str): string of column name to use as target; must appear as a column in df
            pos_label (str/int): string/int of label to use as the positive label (i.e. what the checklist predicts)
            col_subset (list): if df contains extra columns, col_subset is a list of str with columns for modelling
            percentile_inc (int/float): for numeric variables, the percentile increment to binarize by
            op (char/dict[str:char]): for ordinal and numeric variables, the direction of binarized features
            already_binarized: features and target in df are pre-binarized
            thresholds (dict[str:list]): manually provide thresholds for each feature
            add_complement (bool/dict[str:bool]): adds the complement of all generated features
            use_optbin (bool): compute thresholds using the optbinning package
            max_n_bins (int): maximum number of bins for continuous features when computing thresholds with optbinning
            complement_includes_missing (bool): whether the complement of a column will be positive if the feature is missing
            use_missing_features (bool): whether to include whether a particular feature is missing as a variable
            protected_attrs (list[str]): names of columns to use as protected attributes
            categorical_bin (str): one of ['all', 'one']. If all, will use each value as thresholds. Otherwise, will use most frequent bin.
        '''
        # set fields
        assert(target_name in df)        
        assert(categorical_bin in ['all', 'one'])
        self.target_name = target_name
        self.pos_label = pos_label
        self.col_subset = col_subset
        self.df = df if col_subset is None else df[[target_name] + col_subset]
        self.percentile_inc = percentile_inc
        self.op = op
        self.already_binarized = already_binarized
        self.thresholds = thresholds
        self.add_complements = add_complements
        self.use_optbin = use_optbin
        self.max_n_bins = max_n_bins
        self.complement_includes_missing = complement_includes_missing
        self.use_missing_features = use_missing_features     
        self.protected_attrs = protected_attrs
        self.categorical_bin = categorical_bin
                
        if protected_attrs:
            self.protected_groups = df[protected_attrs]        

        if self.already_binarized:
            self.target = self._binarize_target(self.df[self.target_name])

            # check df
            self.binarized_df = self.df.drop(columns = [target_name])
            for i in self.binarized_df:
                if i != target_name:
                    assert ((~self.binarized_df[i].isin([0, 1])).sum()) == 0, f'{i} not binary!'

            # initialize other required fields
            self.col_types = {i: 'prebinarized' for i in self.binarized_df}
            self.thresholds = None
            self.feature_mapping = {i: Feature(i, [i], 'prebinarized') for i in self.binarized_df}

        else:
            # binarizes target
            self.target = self._binarize_target(self.df[self.target_name])

            # binarizes df
            self._binarize_df()

        # remove columns with the same data
        before_removal_columns = self.binarized_df.columns
        self.remaining_features = list(self.binarized_df.T.drop_duplicates().T.columns)

        # remove columns where all samples are the same value
        i=0
        while i < len(self.remaining_features):
            if len(self.binarized_df[self.remaining_features[i]].unique()) == 1:
                del self.remaining_features[i]
            else:
                i += 1

        self.binarized_df = self.binarized_df[self.remaining_features]
        log.info("Removed %s non-informative columns: %s" % (len(before_removal_columns) - len(self.remaining_features),
                                                                set(before_removal_columns) - set(self.remaining_features)))
        log.info(f"Binary dataframe: {self.binarized_df.shape[1]} binary features and {self.binarized_df.shape[0]} samples")

    def _binarize_target(self, target_arr):
        assert(len(target_arr.unique()) == 2), "More than two unique target values!"
        assert(self.pos_label in target_arr.unique()), "Positive label not present in target column!"
        target = (target_arr == self.pos_label).astype(int)
        target[target == 0] = -1
        return target

    def _binarize_df(self):
        df = self.df.copy()
        df = df.drop(columns = [self.target_name])
        self.pctle_list = np.arange(self.percentile_inc, 100, self.percentile_inc)
        self.col_types = infer_types(df)
        if self.thresholds is None:
            if self.use_optbin:
                self.thresholds = compute_optbin_thresholds(df, self.col_types, self.target, self.max_n_bins)
            else: # build thresholds from percentiles
                self.thresholds = {}                
                for col in df:
                    binarized_names = []
                    if self.col_types[col] == 'categorical':
                        if self.categorical_bin == 'all':
                            self.thresholds[col] = df[col].unique()
                        elif self.categorical_bin == 'one':
                            self.thresholds[col] = [df[col].value_counts().sort_values(ascending = False).index[0]]
                        if len(self.thresholds[col]) > df.shape[0] // 2:
                            log.warning(f'categorical column "{col}" has many ({len(self.thresholds[col])}) unique values')
                    else:
                        if self.col_types[col] == 'ordinal':
                            if self.categorical_bin == 'all':
                                split_vals = np.sort(df[col].unique())
                            elif self.categorical_bin == 'one':
                                split_vals = [df[col].dropna().median()]
                        elif self.col_types[col] == 'numeric':
                            split_vals = np.sort(pd.unique(np.nanpercentile(df[col].astype(float).values, q = self.pctle_list)))
                            split_vals = split_vals[~np.isnan(split_vals)].tolist()
                            if pd.isnull(df[col]).sum() > 0:
                                split_vals.append(np.nan)
                        else:
                            raise NotImplementedError

                        self.thresholds[col] = split_vals

        df, self.feature_mapping = self._apply_existing_thresholds(df)
        self.binarized_df = df.drop(columns = list(self.col_types.keys()))

    def _apply_existing_thresholds(self, df):
        feature_mapping = {}
        for col in df:
            binarized_names = []
            complement_pairs = []
            add_complements_i = self.add_complements if isinstance(self.add_complements, bool) else self.add_complements[col]
            op_i = self.op if isinstance(self.op, str) else self.op[col]
            if self.col_types[col] == 'categorical':
                for val in self.thresholds[col]:
                    if pd.isnull(val):
                        if not self.use_missing_features:
                            continue
                        name = col+"_Missing"
                        df[name] = pd.isnull(df[col]).astype(int)
                        comp_name = col+"_NotMissing"
                    else:
                        name = f'{col}=={val}'
                        df[name] = (df[col] == val).astype(int)
                        comp_name = f'{col}!={val}'
                    binarized_names.append(name)

                    if add_complements_i:
                        binarized_names.append(comp_name)
                        
                        if self.complement_includes_missing:                        
                            df[comp_name]= (df[name] == 0).astype(int)
                        else:
                            df[comp_name] = ((df[col] != val) & (~pd.isnull(df[col]))).astype(int)
                            
                        complement_pairs.append((name, comp_name))

                # only 2 categories, can potentially use other category as compl
                if add_complements_i and len(self.thresholds[col]) == 2:
                    n1, c1, n2, c2 = binarized_names[-4:]
                    if (df[n1] == df[c2]).all():
                        complement_pairs = complement_pairs[:-2]
                        complement_pairs.append((n1, n2))
                        binarized_names = binarized_names[:-4] + [n1, n2]
                        df = df.drop(columns = [c1, c2])

            else:
                split_vals = self.thresholds[col]
                for c, val in enumerate(split_vals):
                    if pd.isnull(val):
                        if not self.use_missing_features:
                            continue
                        name = col+"_Missing"
                        df[name] = pd.isnull(df[col]).astype(int)
                        comp_name = col+"_NotMissing"
                    else:
                        name = f'{col}{op_i}{val}'
                        comp_name = f'{col}{inverse_ops[op_i]}{"~" if self.complement_includes_missing else ""}{val}'
                        df[name] = compr(df[col], val, op_i).fillna(0).astype(int)

                    binarized_names.append(name)
                    if add_complements_i:
                        binarized_names.append(comp_name)
                        if self.complement_includes_missing:   
                            df[comp_name]= (df[name] == 0).astype(int)
                        else:
                            df[comp_name]= compr(df[col], val, inverse_ops[op_i]).fillna(0).astype(int)
                        complement_pairs.append((name, comp_name))

            feature_mapping[col] = Feature(col, binarized_names, dtype = self.col_types[col], complement_pairs = complement_pairs)

        return df, feature_mapping

    def subset_by_protected_group(self, attribute_name, grp_name):
        mask = self.protected_groups[attribute_name] == grp_name
        return self.binarized_df[mask], self.target[mask]    

    def shuffle_with_replacement(self, stratify_by = None):
        if stratify_by is None:
            idxs = np.random.choice(np.arange(len(self.binarized_df)), size = len(self.binarized_df))
        else:
            protected_groups_with_idx = self.protected_groups[[stratify_by]].assign(idx = np.arange(len(self.protected_groups)))
            idxs = protected_groups_with_idx.groupby(stratify_by).apply(lambda x: x.sample(n = len(x), replace = True)).sample(frac=1)['idx'].values
        
        new_ds = copy.deepcopy(self)
        new_ds.df = self.df.iloc[idxs]
        new_ds.binarized_df = self.binarized_df.iloc[idxs]
        new_ds.target = self.target.iloc[idxs]
        new_ds.protected_groups = self.protected_groups.iloc[idxs]
        return new_ds

    def apply_transform(self, df_test):    
        if self.already_binarized:
            # log.warn("apply_transform called with already_binarized = True")
            target = self._binarize_target(df_test[self.target_name])
            df_new = df_test.drop(columns = [self.target_name])
        else:
            df_new = df_test.copy()
            df_new = df_new[self.df.columns]

            # binarize target
            target = self._binarize_target(df_new[self.target_name])

            # apply threshold to df
            df_new = df_new.drop(columns = [self.target_name])
            df_new, _ = self._apply_existing_thresholds(df_new)

            df_new = df_new.drop(columns = list(self.col_types.keys()))

        # create returned BinaryDataset
        new_ds = BinaryDataset.__new__(BinaryDataset)
        new_ds.df = df_test
        new_ds.target_name = self.target_name
        new_ds.pos_label = self.pos_label
        new_ds.col_subset = self.col_subset
        new_ds.percentile_inc = self.percentile_inc
        new_ds.op = self.op
        new_ds.already_binarized = self.already_binarized
        new_ds.binarized_df = df_new[self.remaining_features]
        new_ds.target = target
        new_ds.col_types = self.col_types
        new_ds.thresholds = self.thresholds
        new_ds.feature_mapping = self.feature_mapping
        new_ds.add_complements = self.add_complements
        new_ds.complement_includes_missing = self.complement_includes_missing
        new_ds.remaining_features = self.remaining_features
        new_ds.use_missing_features = self.use_missing_features
        new_ds.protected_attrs = self.protected_attrs
        new_ds.categorical_bin = self.categorical_bin
        
        if self.protected_attrs:
            new_ds.protected_groups = df_test[self.protected_attrs]   

        return new_ds

    def __len__(self):
        return self.binarized_df.shape[0]




import tempfile
import pandas as pd
try:
    import cplex
    from cplex import Cplex
    from cplex.exceptions import CplexError
except:
    print("CPLEX package not installed. Please download IBM ILOG CPLEX Optimization Studio here " + 
    'and then install the provided python package: https://www.ibm.com/products/ilog-cplex-optimization-studio')
    raise

import numpy as np
from IPChecklists.utils import set_seed, StatsCallback, binary_metrics, ProcessedDataset
from IPChecklists.constraints_cplex import *
print_vnames = lambda vfmt, vcnt: list(map(lambda v: vfmt % v, range(vcnt)))
from IPChecklists.Covers import CoverResult
from IPChecklists.Checklist import Checklist

print("Using CPLEX version " + cplex.__version__)

logging.basicConfig(level='INFO')
log = logging.getLogger()


def _add_variable(model, name, obj, lb, ub, vtype):

    name = [name] if type(name) is not list else name
    obj = [float(obj)] if type(obj) is not list else obj
    ub = [float(ub)] if type(ub) is not list else ub
    lb = [float(lb)] if type(lb) is not list else lb
    vtype = [vtype] if type(vtype) is not list else vtype

    assert isinstance(name[0], str)
    assert isinstance(obj[0], float)
    assert isinstance(ub[0], float)
    assert isinstance(lb[0], float)
    assert isinstance(vtype[0], str)

    model.variables.add(names = name, obj = obj, lb = lb, ub = ub, types = vtype)

def _add_constraint(model, lhs_ind, rhs_ind, sense = 'E', name = None):
    # sense = "G" -> lhs >= rhs

    if not isinstance(lhs_ind, list):
        lhs_ind = [lhs_ind]

    if not isinstance(rhs_ind, list):
        rhs_ind = [rhs_ind]

    assert len(lhs_ind) > 0 and len(rhs_ind) > 0

    con_args = {
        'senses': [sense],
        'lin_expr': [cplex.SparsePair(ind = lhs_ind + rhs_ind, val = [1.0] * len(lhs_ind) + [-1.0] * len(rhs_ind))],
        'rhs': [0.0],
        }

    if name is not None:
        con_args['names'] = [name]

    model.linear_constraints.add(**con_args)


class ChecklistMIP(object):
    def __init__(self, dataset, model_name = 'checklist', cost_func = '01',  one_feature_per_group = True,
                    lplus_weight = None, lminus_weight = None, compress = True):
        '''
        dataset: object of type BinaryDataset  
        cost_func: one of ['01', 'FNR', 'FPR'] 
        compress: whether to compress samples that have the same features. Must be False for fairness constraints.
        '''
        self.dataset = dataset
        target = dataset.target.copy()
        target[target == 0] = -1                    
        self.binarized_df = dataset.binarized_df.copy() 
        self.remaining_features = dataset.remaining_features
        
        self.target = target.values
        self.model_name = model_name
        self.cost_func = cost_func
        self.constraints = [] # named constraints
        self.solved = False     

        self.prod = self._init(compress = compress) # processed dataset

        if cost_func == '01':
            self.lplus_weight = 1.0
            self.lminus_weight = 1.0

        elif cost_func == 'FPR': # minimize false positives = minimize errors on negative samples
            self.lplus_weight = 1.0 
            self.lminus_weight = self.prod.n_samples_minus + 0.5

        elif cost_func == 'FNR': # 
            self.lplus_weight = self.prod.n_samples_plus + 0.5
            self.lminus_weight = 1.0 
        else:
            raise NotImplementedError

        self.lplus_weight = lplus_weight if lplus_weight is not None else self.lplus_weight      
        self.lminus_weight = lminus_weight if lminus_weight is not None else self.lminus_weight
        
        self.regularizer1 = min(self.lplus_weight, self.lminus_weight) * 0.95 / float(self.binarized_df.shape[0]) / float(len(self.dataset.feature_mapping))
        self.regularizer2 = min(self.lplus_weight, self.lminus_weight) * 0.95 * self.regularizer1 / float(len(self.dataset.feature_mapping))
        self.one_feature_per_group = one_feature_per_group    

    def _init(self, compress = True):
        # compress data
        df = pd.DataFrame(self.binarized_df)
        df['target'] = self.target
        if compress:
            log.info("Before compression: %s rows" % df.shape[0])        
            df = df.groupby(df.columns.tolist()).size().reset_index(name='weight_i')
            log.info("After compression: %s rows" % df.shape[0])
        else:
            df = df.reset_index(drop = True)
            df['weight_i'] = 1

        n = df['weight_i'].values
        target = df['target'].values
        I_plus = df.query('target == 1').index.tolist()
        I_minus = df.query('target == -1').index.tolist()
        df = df.drop(columns = ['target', 'weight_i'])
        mat = df.values

        return ProcessedDataset(mat, target, I_plus, I_minus, n)

    def build_problem(self, N_constraint = None, M_constraint = None, use_indicator = False):  
        (n_features, n_samples, n_samples_plus, n_samples_minus, n, I_plus, I_minus, mat) = (self.prod.n_features, 
            self.prod.n_samples, self.prod.n_samples_plus, self.prod.n_samples_minus, self.prod.n, self.prod.I_plus, self.prod.I_minus, self.prod.mat)

        if N_constraint is not None:
            assert isinstance(N_constraint, MaxNumFeatureConstraint)
            N_min, N_max = N_constraint.get_min_max(n_features)
        else:
            N_min, N_max = 1.0, float(n_features)
            
        if M_constraint is not None:
            assert isinstance(M_constraint, MConstraint)
            M_min, M_max = M_constraint.get_min_max(n_features)
        else:
            M_min, M_max = 0.0, float(n_features)        
        
        M_max = min(M_max, N_max)

        model = Cplex()
        model.objective.set_sense(model.objective.sense.minimize)
        constraints = model.linear_constraints
        #self.add_constraint(LossConstraint('<=', min(len(I_plus), len(I_minus))))
        names = {
            'lamb': print_vnames('lamb%d', n_features), # 1 if checklist includes item j
            'z': print_vnames('z%d', n_samples), # 1 if mistake on sample i is predicted correctly
            'M': 'M',
            'N': 'N',
            'lplus': 'lplus', # errors on positive samples
            'lminus': 'lminus', # errors on negative samples
            'l': 'l'
            }
            
        # decision variables
        _add_variable(model, names['lamb'], [0.0] * n_features, [0.0] * n_features, [1.0] * n_features, ['B'] * n_features)
        _add_variable(model, names['M'], [self.regularizer2], [M_min], [M_max], ['I'])
        _add_variable(model, names['z'], [0.0] * n_samples, [0.0] * n_samples, [1.0] * n_samples, ['B'] * n_samples)

        # number of rules
        _add_variable(model, names['N'], [self.regularizer1], [N_min], [N_max], ['I'])
        _add_constraint(model, names['N'], names['lamb'], sense = 'E')

        # M < N
        _add_constraint(model, names['M'], names['N'], sense = 'L')

        # number of mistakes on positive/negative samples
        _add_variable(model, names['lplus'], [self.lplus_weight], [0.0], [float(n_samples_plus)], ['I']) # false negatives
        _add_variable(model, names['lminus'], [self.lminus_weight], [0.0], [float(n_samples_minus)], ['I']) # false positives
        _add_variable(model, names['l'], [0.0], [0.0], [float(min(n_samples_plus, n_samples_minus))], ['I'])

        constraints.add(lin_expr = [cplex.SparsePair(ind = [names['lplus']] + ['z%d' % i for i in I_plus], val = [1.0] + (-n[I_plus]).tolist())], senses = ['E'], rhs = [0.0])
        constraints.add(lin_expr = [cplex.SparsePair(ind = [names['lminus']] + ['z%d' % i for i in I_minus], val = [1.0] + (-n[I_minus]).tolist())], senses = ['E'], rhs = [0.0])
        _add_constraint(model, names['l'], [names['lplus'], names['lminus']], sense = 'E')
        
        if use_indicator:
            # positive samples: z[i] = 0 -> M - num_checked <= 0   
            for i in I_plus:
                model.indicator_constraints.add(
                    lin_expr = cplex.SparsePair(ind = [names['M']] + names['lamb'], val = [1.0] + (-mat[i, :]).tolist() ),
                    sense = 'L',
                    rhs = 0.0,
                    indvar = names['z'][i],
                    complemented = True                    
                )
            # negative examples: z[i] = 0 -> num_checked - M +1 <= 0
            for i in I_minus:
                model.indicator_constraints.add(
                    lin_expr = cplex.SparsePair(ind = [names['M']] + names['lamb'],  val = [-1.0] + mat[i, :].tolist()),
                    sense = 'L',
                    rhs = -1.0,
                    indvar = names['z'][i],
                    complemented = True                    
                )
                
        else:
            # compute B
            B = np.zeros(n_samples)
            B[I_plus] = M_max + 1 #todo: this can be tighter I think... work from max(M - \sum(lambda[j]x[i,j]) over M & lambda.
            B[I_minus] = np.minimum(mat[I_minus, :].sum(axis = 1), N_max) - M_min + 2
            B = B.tolist()

            # z_i = 1 if mistake on negative example
            for i in I_minus:
                # B[i]*z[i] + M - num_checked >= 1
                constraints.add(lin_expr = [cplex.SparsePair(ind = [names['z'][i], names['M']] + names['lamb'], val = [B[i], 1.0] + (-mat[i, :]).tolist())],
                                senses = ['G'],
                                rhs = [1.0])

            # z_i = 1 if mistake on positive example
            for i in I_plus:
                # B[i]*z[i] - M + num_checked >= 0       
                constraints.add(lin_expr = [cplex.SparsePair(ind = [names['z'][i], names['M']] + names['lamb'], val = [B[i], -1.0] + mat[i, :].tolist())],
                                senses = ['G'],
                                rhs = [0.0])


        # constraints
        for con in self.constraints:
            con_formulation = con.add(names, self.dataset, self.binarized_df, )
            if 'SOS' in con_formulation:
                model.SOS.add(type = "1", **con_formulation)
            else:           
                model.linear_constraints.add(**con_formulation)

        if self.one_feature_per_group: # add one-feature-per-group constraints
            for ft_name in self.dataset.feature_mapping:
                con = FeaturesPerGroupConstraint(ft_name, '<=', 1)
                con_formulation = con.add(names, self.dataset, self.binarized_df, )
                model.SOS.add(type = "1", **con_formulation)
                
        # add complement-pair constraints
        for ft_name in self.dataset.feature_mapping:
            for n1, n2 in self.dataset.feature_mapping[ft_name].complement_pairs:
                con = MostOneofTwoConstraint(n1, n2)
                con_formulation = con.add(names, self.dataset, self.binarized_df)
                if con_formulation:
                    model.SOS.add(type = "1", **con_formulation)
                
        # add the two constraints passed to this function to named constraints
        if M_constraint is not None:
            self.add_constraint(M_constraint)
            
        if N_constraint is not None:
            self.add_constraint(N_constraint)

        self.model = model
        self.names = names
        self.N_min, self.N_max, self.M_min, self.M_max = N_min, N_max, M_min, M_max

    def add_constraint(self, constraint):
        assert isinstance(constraint, Constraint)
        self.constraints.append(constraint)

    def solve(self, max_seconds = None, threads = 1, return_incumbents = False,
              display_progress = False, emphasis = 1, variableselect = 3,
              memory_limit = 1024*20, seed = 42):

        if not hasattr(self, 'model'):
            log.warn("Must build problem first!")

        if self.solved:
            log.warn('problem is already solved!')
            return

        # parameters
        set_seed(seed)
        self.model.parameters.randomseed.set(seed)
        self.model.parameters.mip.strategy.file = 3 #nodefile compressed
        self.model.parameters.workdir = tempfile.gettempdir()
        self.model.parameters.emphasis.mip = emphasis
        self.model.parameters.mip.strategy.variableselect = variableselect
        self.model.parameters.mip.limits.treememory = memory_limit # https://www.ibm.com/support/pages/apar/RS03251

        if max_seconds is not None:
            self.model.parameters.timelimit.set(max_seconds)

        self.model.parameters.threads.set(threads)
        if display_progress == False:
            self.model.parameters.mip.display.set(0)
            self.model.set_results_stream(None)
            self.model.set_log_stream(None)

        sol_idx = self.model.variables.get_indices(self.names['lamb'] + [self.names['M']])
        self._add_stats_callback(sol_idx, store_solutions = return_incumbents)
        self.model.solve()

        solution_info = self.get_solution_info()
        if solution_info['iterations'] > 0:
            progress_info, progress_incumbents = self._stats_callback.get_stats()
        else:
            progress_info, progress_incumbents = [], []

        if solution_info['has_solution']:
            self.solved = True
            print(f"Found solution with objective {solution_info['objval']} and optimality gap {solution_info['gap']:.2%}.")
            self.lamb = np.array(self.model.solution.get_values(self.names['lamb']))
            self.M = self.model.solution.get_values(self.names['M'])
            self.N = self.model.solution.get_values(self.names['N'])
        else:
            log.warn('No feasible solution found')

        return solution_info, progress_info, progress_incumbents


    def _add_stats_callback(self, sol_idx, store_solutions = False):
        if not hasattr(self, '_stats_callback'):
            min_idx = min(sol_idx)
            max_idx = max(sol_idx)
            assert np.array_equal(np.array(sol_idx), np.arange(min_idx, max_idx + 1))

            cb = self.model.register_callback(StatsCallback)
            cb.initialize(store_solutions, solution_start_idx = min_idx, solution_end_idx = max_idx)
            self._stats_callback = cb

    def get_solution_info(self):
        """returns information associated with the current best solution for the mip"""
        INITIAL_SOLUTION_INFO = {
            'status': 'no solution exists',
            'status_code': float('nan'),
            'has_solution': False,
            'has_mipstats': False,
            'iterations': 0,
            'nodes_processed': 0,
            'nodes_remaining': 0,
            'values': float('nan'),
            'objval': float('nan'),
            'upperbound': float('nan'),
            'lowerbound': float('nan'),
            'gap': float('nan'),
            }

        info = dict(INITIAL_SOLUTION_INFO)
        try:
            sol = self.model.solution
            progress_info = {'status': sol.get_status_string(),
                             'status_code': sol.get_status(),
                             'iterations': sol.progress.get_num_iterations(),
                             'nodes_processed': sol.progress.get_num_nodes_processed(),
                             'nodes_remaining': sol.progress.get_num_nodes_remaining()}
            info.update(progress_info)
            info['has_mipstats'] = True
        except CplexError:
            pass

        try:
            sol = self.model.solution
            solution_info = {'values': np.array(sol.get_values()),
                             'objval': sol.get_objective_value(),
                             'upperbound': sol.MIP.get_cutoff(),
                             'lowerbound': sol.MIP.get_best_objective(),
                             'gap': sol.MIP.get_mip_relative_gap()}
            info.update(solution_info)
            info['has_solution'] = True
        except CplexError:
            pass

        return info

    def add_initial_solution_from_mip(self, other_mip):
        assert(isinstance(other_mip, ChecklistMIP))
        var_names = other_mip.model.variables.get_names()
        var_values = other_mip.model.solution.get_values(var_names)
        self.model.MIP_starts.add([var_names, var_values], self.model.MIP_starts.effort_level.check_feasibility)    
        
    def add_initial_solution_from_checklist(self, checklist):
        assert(isinstance(checklist, Checklist))
        for i in checklist.column_names:
            if i not in self.remaining_features:
                log.warn(f'Trying to add initial solution with different feature: {i}. Skipping.')
                return None
        self.add_initial_solution_from_weights(checklist.lamb, checklist.M) 
        
    def add_initial_solution_from_weights(self, lamb, M):
        assert(len(lamb) == len(self.names['lamb']))
        assert(M <= self.M_max and M >= self.M_min), (M, self.M_max, self.M_min)
        assert(i in [0, 1, 0., 1.] for i in lamb)
        N_init = sum(lamb)
        assert(N_init <= self.N_max and N_init >= self.N_min), (N_init, self.N_max, self.N_min)
        
        var_names = self.names['lamb'] + [self.names['M'], self.names['N']]
        var_values = [float(i) for i in lamb] + [float(M), float(N_init)]
        self.model.MIP_starts.add([var_names, var_values], self.model.MIP_starts.effort_level.solve_MIP)   

    def pretty_print_soln(self):
        string = ''
        string += ('-' * 30) + '\n'
        string += ('Model name: ' + self.model_name ) + '\n'
        string += ('Regularizer1: ' + str(self.regularizer1) ) + '\n'
        string += ('Regularizer2: ' + str(self.regularizer2) ) + '\n'
        string += ('Constraints: ' + "None" + '\n' if len(self.constraints) == 0 else '\n'+ '\n'.join(str(i) for i in self.constraints) + '\n')
        string += ('-' * 30) + '\n'
        string += ('Training loss: %s' % (self.model.solution.get_objective_value())) + '\n'
        string += ('Num errors: %s' % self.model.solution.get_values(self.names['l'])) + '\n'
        string += (f"Training optimality gap: {self.model.solution.MIP.get_mip_relative_gap():.2%}") + '\n'
        string += ('-' * 30) + '\n'
        for v, col in zip(self.lamb, self.binarized_df.columns):
            if v == 1:
                string += (col) + '\n'
                
        string += '\n'
        string += ('M = %s, N = %s'%(self.M,  self.N)) + '\n'
        string += ('-' * 30)+ '\n'
        return string   
                
    def get_progress_logs(self):
        return self._stats_callback.get_stats()[0]

    def plot_progress_logs(self):
        plogs = self.get_progress_logs()
        import matplotlib.pyplot as plt
        
        fig, axs = plt.subplots(1, 2)
        ax = axs[0]
        ax.plot(plogs['runtime'], plogs['lowerbound'], '-', label = 'lb')
        ax.plot(plogs['runtime'], plogs['upperbound'], '-', label = 'ub')
        ax.legend()
        ax.set_xlabel('t (s)')
        ax.set_ylabel('Objective')
                
        ax = axs[1]
        ax.plot(plogs['nodes_processed'], plogs['lowerbound'], '-', label = 'lb')
        ax.plot(plogs['nodes_processed'], plogs['upperbound'], '-', label = 'ub')
        ax.legend()
        ax.set_xlabel('# Nodes')
        ax.set_ylabel('Objective')
        
        plt.tight_layout()
        plt.show()
        
    def to_checklist(self):
        return Checklist(from_mip = True, mip = self)


from IPChecklists.Checklist import Checklist
import numpy as np
from pathlib import Path
import pickle
import logging
logging.basicConfig(level='INFO')
log = logging.getLogger()

def checklists_from_lr(lr_model, train_ds, test_ds = None, use_basic = False):
    if not train_ds.add_complements:
        log.warn("Dataset does not contain complements. Unit weighting might give poor performance " +
                "under the presence of negative parameters.")
    coefs = lr_model.steps[-1][1].coef_[0]
    weights = np.copy(coefs)
    abs_weights = np.abs(weights)
    
    if not use_basic:
        for i in train_ds.feature_mapping: # set all features in one group that do not have largest params to zero
            inds = []
            for j in train_ds.feature_mapping[i].binarized_names:
                try:
                    inds.append(list(train_ds.binarized_df.columns).index(j))
                except ValueError:
                    pass
            if len(inds) == 0:
                continue
            max_abs = np.max(abs_weights[np.array(inds)])
            for j in inds:
                if np.abs(weights[j]) < max_abs:
                    weights[j] = 0
     
    weights[weights > 0] = 1
    # if a weight is negative, take the complement feature as 1
    for i in train_ds.feature_mapping:
        for j in train_ds.feature_mapping[i].binarized_names:
            try:
                cur_ind = list(train_ds.binarized_df.columns).index(j)
            except ValueError:
                continue
            if weights[cur_ind] < 0 and train_ds.add_complements:         
                try:
                    oppo_names = [k for k in train_ds.feature_mapping[i].complement_pairs if j in k][0]
                    oppo_name = oppo_names[1 - oppo_names.index(j)]
                    oppo_ind = list(train_ds.binarized_df.columns).index(oppo_name)
                    weights[oppo_ind] = 1
                    weights[cur_ind] = 0    
                except (ValueError, IndexError) as e: # complement feature was dropped from df due to redundancy, or never existed
                    log.warn(f"Complement feature to {j} not found.")
                    weights[cur_ind] = 0    
                
    weights[weights < 0] = 0
    N = (weights == 1).sum()     
    checklists = []
    train_metrics, test_metrics = [], []
    for M in range(1, N+1):
        checklists.append(Checklist(from_mip = False, from_weights = True, lamb = weights,
                                         M = M, column_names = list(train_ds.binarized_df.columns)))
        train_metrics.append(checklists[-1].get_metrics(train_ds))
        if test_ds is not None:
            test_metrics.append(checklists[-1].get_metrics(test_ds))
        
    return checklists, train_metrics, test_metrics, N



import numpy as np
import pandas as pd
from pathlib import Path
import pickle
import json
from tqdm import tqdm

class SolutionPool():
    def __init__(self, dataset_name, binarization):
        self.dataset_name = dataset_name
        self.binarization = binarization
        self.pool = []
        self.df = pd.DataFrame(columns = ['ind', 'M', 'N', 'origin'])

    def __len__(self):
        return len(self.pool)

    def add_from_dir(self, dir, origin = None, enforce_fold = None):
        '''
        Adds a directory of trained checklist to the pool, with a given origin (str)
        '''
        for i in Path(dir).glob('**/checklists'):
            args = json.load((i.parent/'args.json').open('r'))
            if enforce_fold is not None and args['fold'] != enforce_fold:
                continue
            if args['dataset'] == self.dataset_name and args['binarization'] == self.binarization:
                checklists = pickle.load(i.open('rb'))
                for checklist in checklists:
                    self.add_to_pool(checklist, origin if origin is not None else args['experiment_name'])

        for i in Path(dir).glob('**/checklist'):
            args = json.load((i.parent/'args.json').open('r'))
            if enforce_fold is not None and args['fold'] != enforce_fold:
                continue
            if args['dataset'] == self.dataset_name and args['binarization'] == self.binarization:
                checklist = pickle.load(i.open('rb'))
                self.add_to_pool(checklist, origin if origin is not None else args['experiment_name'])

    def add_to_pool(self, checklist, origin = None):
        '''
        Adds a single checklist to the pool, with potentially a specified origin (str)
        '''        
        row = {
            'ind': len(self.pool),
            'M': checklist.M,
            'N': checklist.N,
            'origin': origin
        }
        self.pool.append(checklist)
        self.df = self.df.append(row, ignore_index = True)

    def get_subset(self, N = None, N_op = '<=', M = None, M_op = '<=', origin = None):
        '''
        Given a set of constraints, returns a list of Checklist objects that satisfies these constraints
        '''
        self.subset = self.df.copy().query("M >= 1.0")
        if N is not None:
            self.subset = self.subset.query(f"N {N_op} {N}")
        if M is not None:
            self.subset = self.subset.query(f"M {N_op} {M}")
        if origin is not None:
            self.subset = self.subset[self.subset.origin == origin]
        inds = self.subset.ind.values.tolist()
        return [self.pool[c] for c in inds]
        



import numpy as np
import pandas as pd
import logging
logging.basicConfig(level='INFO')
log = logging.getLogger()
from IPChecklists.utils import binary_metrics
import random
import re

class Checklist(object):
    def __init__(self, from_mip = False, from_weights = False, mip = None, lamb = None, M = 0, column_names = []):
        self.from_mip = from_mip
        if from_mip:
            if mip.solved:          
                self.is_valid = True
                self.lamb = mip.lamb.round()
                self.M = mip.M
                self.N = self.lamb.sum()
                self.column_names = mip.remaining_features
                
                # only for mips
                self.progress_log = mip.get_progress_logs()
                self.checklist_str = mip.pretty_print_soln()
                self.solution_info = mip.get_solution_info()                
            else:
                self.is_valid = False
                log.warn("MIP with no solution passed into Checklist constructor!")
        elif from_weights:
            self.lamb = np.array(lamb)
            self.M = M        
            self.is_valid = True
            self.N = int(self.lamb.sum())
            self.column_names = column_names
        else:
            raise ValueError('No valid input data passed!')    

    def get_num_checked(self, dataset):
        if not self.is_valid:
            raise ValueError("Prediction with invalid checklist!")      
        df_bin = dataset.binarized_df
        if self.from_mip:
            mat = df_bin[self.column_names].values
        else:
            mat = df_bin.values
        return mat @ self.lamb

    def predict(self, dataset):
        return self.get_num_checked(dataset) >= self.M    
    
    def get_metrics(self, ds):
        target = np.copy(ds.target)
        target[target == -1] = 0
        
        preds = self.predict(ds)
        res = binary_metrics(target, preds)
        return res    

    def get_metrics_from_df(self, df_bin, target):
        target = np.copy(target)
        target[target == -1] = 0
        if self.from_mip:
            mat = df_bin[self.column_names].values
        else:
            mat = df_bin.values

        preds = (mat @ self.lamb) >= self.M    
        res = binary_metrics(target, preds)
        return res   

    def get_fairness_metrics_simple(self, ds, attributes):
        rows = {}
        for attr in attributes:        
            for grp in ds.protected_groups[attr].unique():
                df_bin_i, target_i = ds.subset_by_protected_group(attr, grp)
                metrics_i = self.get_metrics_from_df(df_bin_i, target_i)
                rows[grp] = metrics_i
        return pd.DataFrame(rows)
    
    def __repr__(self):
        return self.pretty_print_soln()

    def pretty_print_soln(self):
        string = ''
        for i in range(len(self.lamb)):
            if self.lamb[i] == 1:
                string += (self.column_names[i]) + '\n'
        string += '\n'
        string += ('M = %s, N = %s'%(self.M,  self.N)) + '\n'
        return string

    def get_selected_cols(self):
        return [self.column_names[c] for c, i in enumerate(self.lamb) if i == 1]
    
    def __eq__(self, other):
        return self.lamb == other.lamb and self.M == other.lamb and self.column_names == other.column_names

    def to_latex(self, target_name):
        st = r'\begin{tabular}[c]{@{}>{\sffamily}l@{}r}' + '\n'
        st += f'Predict {target_name} if {int(self.M)}+ Items are Checked\\\\\n\\hline\n'
        for col in self.get_selected_cols():
            col = col.replace('==', '=').replace('_', '\_')
            if col.count('=') <= 1 and col.count('~') <= 1:
                col = col.replace('~', '=')
                p = re.compile(r'(?P<compr>[<>!]?=)')
                col = p.sub(' $\g<compr>$ ', col).replace('<=', '\\leq').replace('>=', '\\geq')
            st+= col + r'& $\square$\\' + '\n'
        st += '\end{tabular}\n'
        return st



from setuptools import setup

setup(name='predictive_checklists',
      version='0.1',
      description='Creating Predictive Checklists Using Integer Programming',
      url='https://github.com/MLforHealth/predictive_checklists',
      author='Haoran Zhang',
      author_email='haoranz@mit.edu',
      license='BSD',
      packages=['IPChecklists'],
      python_requires='>=3.7',
      install_requires= [
        'numpy>=1.19.0',
        'pandas>=1.1.5',
        'scikit-learn>=0.24.1',
        'scipy>=1.4.1',
        'matplotlib',
        'tqdm',
        'imbalanced-learn>=0.8.0',
        'xgboost>=0.90',
        'optbinning>=0.8.0',
        'mip>=1.13.0',
        'openpyxl>=3.0.9'
      ],
      include_package_data=True)


## Formulation of the problem using PythonMIP - currently not updated
from mip import Model, BINARY, INTEGER, minimize, xsum, compute_features, features, INF, CBC, GRB
import warnings
warnings.warn("Loaded the Python MIP interface. We provide NO GUARANTEES on the performance of the Python MIP solvers. " + 
        "Where possible, please use CPLEX instead.")
from tqdm import tqdm
import numpy as np
import random
from .utils import set_seed
import pandas as pd
import logging
import os
from IPChecklists.utils import ProcessedDataset
from IPChecklists.constraints_pythonmip import *
from IPChecklists.Checklist import Checklist

logging.basicConfig(level='INFO')
log = logging.getLogger()
    
class ChecklistMIP():
    def __init__(self, dataset, model_name = 'checklist', cost_func = '01', solver = 'CBC',
                    one_feature_per_group = True, lplus_weight = None, lminus_weight = None, compress = True):
        '''
        dataset: object of type BinaryDataset  
        cost_func: one of ['01', 'FNR', 'FPR'] 
        solver: 'CBC' or 'GRB'
        compress: whether to compress samples that have the same features. Must be False for fairness constraints.
        '''
        self.dataset = dataset
        target = dataset.target.copy()
        target[target == 0] = -1                    
        self.binarized_df = dataset.binarized_df.copy() 
        self.remaining_features = dataset.remaining_features
        
        self.target = target.values
        self.model_name = model_name
        self.cost_func = cost_func
        self.constraints = [] # named constraints
        self.solved = False     
        self.solver = solver

        self.prod = self._init(compress = compress) # processed dataset

        if cost_func == '01':
            self.lplus_weight = 1.0
            self.lminus_weight = 1.0

        elif cost_func == 'FPR': # minimize false positives = minimize errors on negative samples
            self.lplus_weight = 1.0 
            self.lminus_weight = self.prod.n_samples_minus + 0.5

        elif cost_func == 'FNR': # 
            self.lplus_weight = self.prod.n_samples_plus + 0.5
            self.lminus_weight = 1.0 
        else:
            raise NotImplementedError

        self.lplus_weight = lplus_weight if lplus_weight is not None else self.lplus_weight      
        self.lminus_weight = lminus_weight if lminus_weight is not None else self.lminus_weight
        
        self.regularizer1 = min(self.lplus_weight, self.lminus_weight) * 0.95 / float(self.binarized_df.shape[0]) / float(len(self.dataset.feature_mapping))
        self.regularizer2 = min(self.lplus_weight, self.lminus_weight) * 0.95 * self.regularizer1 / float(len(self.dataset.feature_mapping))
        self.one_feature_per_group = one_feature_per_group    
    
    def _init(self, compress = True):
        # compress data
        df = pd.DataFrame(self.binarized_df)
        df['target'] = self.target
        if compress:
            log.info("Before compression: %s rows" % df.shape[0])        
            df = df.groupby(df.columns.tolist()).size().reset_index(name='weight_i')
            log.info("After compression: %s rows" % df.shape[0])
        else:
            df = df.reset_index(drop = True)
            df['weight_i'] = 1

        n = df['weight_i'].values
        target = df['target'].values
        I_plus = df.query('target == 1').index.tolist()
        I_minus = df.query('target == -1').index.tolist()
        df = df.drop(columns = ['target', 'weight_i'])
        mat = df.values

        return ProcessedDataset(mat, target, I_plus, I_minus, n)
    
    def build_problem(self, N_constraint = None, M_constraint = None):
        (n_features, n_samples, n_samples_plus, n_samples_minus, n, I_plus, I_minus, mat) = (self.prod.n_features, 
            self.prod.n_samples, self.prod.n_samples_plus, self.prod.n_samples_minus, self.prod.n, self.prod.I_plus, self.prod.I_minus, self.prod.mat)
        
        if N_constraint is not None:
            assert isinstance(N_constraint, MaxNumFeatureConstraint)
            N_min, N_max = N_constraint.get_min_max(n_features)
        else:
            N_min, N_max = 1.0, float(n_features)
            
        if M_constraint is not None:
            assert isinstance(M_constraint, MConstraint)
            M_min, M_max = M_constraint.get_min_max(n_features)
        else:
            M_min, M_max = 0.0, float(n_features)        
        
        M_max = min(M_max, N_max)

        model = Model(self.model_name, solver_name = self.solver)    
        model.store_search_progress_log = True
        model.preprocess = 1
        model.start = []
                
        # decision variables
        lamb = [model.add_var(var_type=BINARY) for i in range(n_features)]
        z = [model.add_var(var_type=BINARY) for i in range(n_samples)]
        M = model.add_var(var_type=INTEGER, lb = 1, ub = n_features)
        
        # number of rules
        N = xsum(lamb)
        model += N <= N_max
        
        # constrain M
        model += M - N <= 0
        model += M <= M_max
                
        # number of mistakes on positive/negative samples
        lplus = xsum(n[i]*z[i] for i in I_plus)
        lminus = xsum(n[i]*z[i] for i in I_minus)
        l = lplus + lminus
        
        B = np.zeros(n_samples)
        B[I_plus] = M_max + 1
        B[I_minus] = np.minimum(mat[I_minus, :].sum(axis = 1), N_max) - M_min + 2
        B = B.tolist()

        # z_i = 1 if mistake on positive example
        for i in I_plus:
            model += ( 
                B[i]*z[i] >= M - xsum(lamb[j]*mat[i, j] for j in range(n_features))                
            )
        
        # z_i = 1 if mistake on negative example
        for i in I_minus:
            model += ( 
                B[i]*z[i] >= xsum(lamb[j]*mat[i, j] for j in range(n_features)) - M + 1
            )
                
        # cost function
        model.objective = minimize(self.lminus_weight*lminus + self.lplus_weight*lplus 
                + self.regularizer1 * N + self.regularizer2 * M)  
         
        # collect variables
        self.m_vars = {
            'model': model,
            'mat': mat,
            'n': n,
            'I_plus': I_plus,
            'I_minus': I_minus,
            'B': B,
            'lamb': lamb,
            'z': z,
            'M': M,
            'N': N,
            'lplus': lplus,
            'lminus': lminus,
            'l': l
        }
        
        # constraints
        for con in self.constraints:
            con.add(self.m_vars, self.dataset, self.binarized_df)         
            
        if self.one_feature_per_group: # add one-feature-per-group constraints
            for ft_name in self.dataset.feature_mapping:
                con = FeaturesPerGroupConstraint(ft_name, '<=', 1)
                con.add(self.m_vars, self.dataset, self.binarized_df)
                
        # add complement-pair constraints
        for ft_name in self.dataset.feature_mapping:
            for n1, n2 in self.dataset.feature_mapping[ft_name].complement_pairs:
                con = MostOneofTwoConstraint(n1, n2)
                con.add(self.m_vars, self.dataset, self.binarized_df)

        # add the two constraints passed to this function to named constraints
        if M_constraint is not None:
            self.add_constraint(M_constraint)
            
        if N_constraint is not None:
            self.add_constraint(N_constraint)

        self.N_min, self.N_max, self.M_min, self.M_max = N_min, N_max, M_min, M_max
    
    def add_constraint(self, constraint):
        assert(isinstance(constraint, Constraint))
        self.constraints.append(constraint)     

    def solve(self, max_seconds = INF, threads = 1, 
        display_progress = False, emphasis = 1, seed = 42):        
        if not hasattr(self, 'm_vars'):
            log.warn("Must build problem first!")            
        
        if self.solved:
            log.warn('problem is already solved!')
            return         
        
        self.m_vars['model'].threads = threads
        self.m_vars['model'].emphasis = emphasis
        self.m_vars['model'].seed = seed
        self.m_vars['model'].verbose = int(display_progress)
        
        self.status = self.m_vars['model'].optimize(max_seconds = max_seconds)
        if self.m_vars['model'].num_solutions:
            self.solved = True
            print(f"Found optimal solution with objective {self.m_vars['model'].objective_value} and optimality gap {self.get_optimality_gap():.2%}.")
            self.lamb = np.array([self.m_vars['lamb'][i].x for i in range(len(self.m_vars['lamb']))])   
            self.M = self.m_vars['M'].x
            self.N = self.m_vars['N'].x
        else:
            log.warn('No feasible solution found')

        return self.get_solution_info(), self.get_progress_logs(), []
    
    def get_optimality_gap(self):
        temp = self.get_progress_logs().iloc[-1]
        if temp['ub'] == 0.0:
            return 1
        else:            
            return (temp['ub'] - temp['lb'])/temp['ub'] 

    def add_initial_solution_from_checklist(self, checklist):
        assert(isinstance(checklist, Checklist))
        for i in checklist.column_names:
            if i not in self.remaining_features:
                log.warn(f'Trying to add initial solution with different feature: {i}. Skipping.')
                return None
        self.add_initial_solution_from_weights(checklist.lamb, checklist.M) 

    def add_initial_solution_from_weights(self, lamb, M):
        assert(len(lamb) == len(self.m_vars['lamb']))
        assert(M <= self.M_max and M >= self.M_min), (M, self.M_max, self.M_min)
        assert(i in [0, 1, 0., 1.] for i in lamb)
        N_init = sum(lamb)
        assert(N_init <= self.N_max and N_init >= self.N_min), (N_init, self.N_max, self.N_min)
        
        self.m_vars['model'].start = [(self.m_vars['M'], M)] + [(a,b) for a,b in zip(self.m_vars['lamb'], lamb)] 
        
    def pretty_print_soln(self):
        string = ''
        string += ('-' * 30) + '\n'
        string += ('Model name: ' + self.model_name ) + '\n'
        string += ('Regularizer1: ' + str(self.regularizer1) ) + '\n'
        string += ('Regularizer2: ' + str(self.regularizer2) ) + '\n'
        string += ('Constraints: ' + "None" + '\n' if len(self.constraints) == 0 else '\n'+ '\n'.join(str(i) for i in self.constraints) + '\n')
        string += ('-' * 30) + '\n'
        string += ('Training loss: %s' % self.m_vars['model'].objective_value) + '\n'
        string += (f"Training optimality gap: {self.get_optimality_gap():.2%}") + '\n'
        string += ('-' * 30) + '\n'
        for i in range(len(self.m_vars['lamb'])):
            if self.m_vars['lamb'][i].x == 1:
                string += (self.binarized_df.columns[i]) + '\n'
                
        string += '\n'
        string += ('M = %s, N = %s'%(self.m_vars['M'].x, self.m_vars['N'].x)) + '\n'
        string += ('-' * 30)+ '\n'
        return string   
    
    def print_properties(self):
        a,b = compute_features(self.m_vars['model']), features()
        for x,y in zip (a,b):
            print('%s: %s'% (y, x))
            
    def get_progress_logs(self):
        temp = pd.DataFrame(self.m_vars['model'].search_progress_log.log, columns = ['t', 'data'])
        temp['lb'], temp['ub'] = temp['data'].str
        return temp.drop(columns = ['data'])

    def get_solution_info(self):
        return {
            'gap': float(self.m_vars['model'].gap),
            'lowerbound': float(self.m_vars['model'].objective_bound),
            'upperbound': float(self.m_vars['model'].objective_value)
        }

    def to_checklist(self):
        return Checklist(from_mip = True, mip = self)


import logging
import numpy as np
from mip import xsum

logging.basicConfig(level='INFO')
log = logging.getLogger()

def get_min_max(op, val, upper_limit):
    if val > upper_limit:
        val = upper_limit
    
    if op == '<=':
        return 1.0, float(val)
    elif op == '==':
        return float(val), float(val)
    elif op == '>=':
        return float(val), float(upper_limit)

class Constraint():
    def __init__(self):
        pass
    
    def add(self, *args, **kwargs):
        pass

class FNRConstraint(Constraint):
    def __init__(self, N_min):
        super().__init__()
        self.N_min = N_min
        
    def add(self, m_vars, dataset, binarized_df):
        m_vars['model'] += (m_vars['lplus'] <= (dataset.target == 1).sum() * self.N_min)
        
    def __str__(self):
        return f"FNR <= {self.N_min}"

class FPRConstraint(Constraint):
    def __init__(self, N_min):
        super().__init__()
        self.N_min = N_min
        
    def add(self, m_vars, dataset, binarized_df):
        m_vars['model'] += (m_vars['lminus'] <= (dataset.target == 0).sum() * self.N_min)
        
    def __str__(self):
        return f"FPR <= {self.N_min}"    

class MaxNumFeatureConstraint(Constraint): # constraint on N
    def __init__(self, op, N_max):
        super().__init__()
        self.op = op
        self.N_max = N_max
    
    def add(self, m_vars, *args, **kwargs):
        if self.op == '<=':
            m_vars['model'] += m_vars['N'] <= self.N_max
        elif self.op == '==':
            m_vars['model'] += m_vars['N'] == self.N_max
        else:
            raise NotImplementedError

    def get_min_max(self, n_features):
        return get_min_max(self.op, self.N_max, n_features)

    def __str__(self):
        return f"Total number of selected attributes {self.op} {self.N_max}."
    
class MConstraint(Constraint): # constraint on M
    def __init__(self, op, val):
        super().__init__()
        self.op = op
        self.val = val
    
    def add(self, m_vars, *args, **kwargs):
        if self.op == '>=':
            m_vars['model'] += (m_vars['M'] >= self.val)
        elif self.op == '<=':
            m_vars['model'] += (m_vars['M'] <= self.val)
        elif self.op == '==':
            m_vars['model'] += (m_vars['M'] == self.val)    
        else:
            raise NotImplementedError

    def get_min_max(self, n_features):
        return get_min_max(self.op, self.val, n_features)

    def __str__(self):
        return f"M {self.op} {self.val}."        
        
class LossConstraint(Constraint): 
    def __init__(self, op = ">=", l = 1):
        self.l = l
        self.op = op
        
    def add(self, m_vars, *args, **kwargs):
        if self.op == '>=':
            m_vars['model'] += (m_vars['l'] >= self.l)
        elif self.op == '<=':
            m_vars['model'] += (m_vars['l'] <= self.l)        
        else:
            raise NotImplementedError    
            
    def __str__(self):
        return f"Total loss {self.op} {self.l}."
        
class FeaturesPerGroupConstraint(Constraint): # select N features from a particular feature group (ex: age)
    def __init__(self, feature, op = "<=", N = 1):
        self.N = N
        self.feature = feature
        self.op = op
        
    def add(self, m_vars, dataset, binarized_df):
        assert(self.feature in dataset.feature_mapping), f"{self.feature} not in original dataframe!"
        inds = []
        for c, i in enumerate(binarized_df.columns):
            if i in dataset.feature_mapping[self.feature].binarized_names:
                inds.append(c)
        if self.op == '<=':
            m_vars['model'] += xsum(m_vars['lamb'][i] for i in inds) <= self.N
        elif self.op == '>=':
            m_vars['model'] += xsum(m_vars['lamb'][i] for i in inds) >= self.N
        elif self.op == '==':
            m_vars['model'] += xsum(m_vars['lamb'][i] for i in inds) == self.N                 
        else:
            raise NotImplementedError
            
        for i in inds[:self.N]:
            m_vars['model'].start.append((m_vars['lamb'][i], 1.0))
                        
    def __str__(self):
        return f"Total number of attributes from group '{self.feature}' {self.op} {self.N}."
    
class MostOneofTwoConstraint(Constraint): # given names of two binarized features, select at most one of the two
    def __init__(self, n1, n2):
        super().__init__()
        self.n1 = n2
        self.n2 = n2
    
    def add(self, m_vars, dataset, binarized_df):
        try:
            ind1 = list(binarized_df.columns).index(self.n1) 
            ind2 = list(binarized_df.columns).index(self.n2) 

            m_vars['model'] += (m_vars['lamb'][ind1] + m_vars['lamb'][ind2] <= 1)
            
        except ValueError: # 1 or more of the features were removed because of duplication
            pass        
        
    def __str__(self):
        return f"One of {self.n1}, {self.n2}"       

class GroupMetricGapConstraintInterface(Constraint):
    def __init__(self, attr, eps, metric = 'FPR'):
        super().__init__()
        self.attr = attr
        self.eps = eps
        assert metric in ['FPR', 'FNR']
        self.metric = metric

    def add(self, m_vars, dataset, binarized_df):
        assert(self.attr in dataset.protected_attrs)
        grp_vals = dataset.protected_groups[self.attr].unique()
        if self.metric == "FPR":
            target_int = -1
        elif self.metric == 'FNR':
            target_int = 1        

        for c, g1 in enumerate(grp_vals):
            m1 = (dataset.protected_groups[self.attr] == g1).values
            n1 = (dataset.target.values == target_int)[m1].sum()
            inds1 = [i for c, i in enumerate(m_vars['z']) if m1[c] and dataset.target.values[c] == target_int]

            for g2 in grp_vals[c+1:]:
                if g1 != g2:
                    m2 = (dataset.protected_groups[self.attr] == g2).values
                    n2 = (dataset.target.values == target_int)[m2].sum()
                    inds2 = [i for c, i in enumerate(m_vars['z']) if m2[c] and dataset.target.values[c] == target_int]                

                    m_vars['model'] += xsum(inds1)/n1 - xsum(inds2)/n2 <= self.eps
                    m_vars['model'] += xsum(inds2)/n2 - xsum(inds1)/n1 <= self.eps
    
    def __str__(self):
        return f"|{self.metric}(G1) - {self.metric}(G2)| <= {self.eps} for all (G1, G2) in {self.attr}"

class GroupFPRGapConstraint(GroupMetricGapConstraintInterface):
    def __init__(self, attr, eps):
        super().__init__(attr, eps, 'FPR')

class GroupFNRGapConstraint(GroupMetricGapConstraintInterface):
    def __init__(self, attr, eps):
        super().__init__(attr, eps, 'FNR')    

class GroupMetricConstraintInterface(Constraint):
    def __init__(self, attr, eps, metric = 'FNR'):
        super().__init__()
        self.attr = attr
        self.eps = eps
        assert metric in ['FPR', 'FNR']
        self.metric = metric

    def add(self, m_vars, dataset, binarized_df):
        assert(self.attr in dataset.protected_attrs)
        assert(len(m_vars['z']) == len(dataset))
        grp_vals = dataset.protected_groups[self.attr].unique()
        if self.metric == "FPR":
            target_int = -1
        elif self.metric == 'FNR':
            target_int = 1  

        for g in grp_vals:
            m = (dataset.protected_groups[self.attr] == g).values            
            n = (dataset.target.values ==  target_int)[m].sum()
            inds = [i for c, i in enumerate(m_vars['z']) if m[c] and dataset.target.values[c] ==  target_int]
            m_vars['model'] += xsum(inds)/n <= self.eps 
    
    def __str__(self):
        return f"{self.metric}(G) <= {self.eps} for all (G) in {self.attr}"

class GroupFNRConstraint(Constraint):
    def __init__(self, attr, eps):
        super().__init__(attr, eps, 'FNR')
        
class GroupFPRConstraint(Constraint):
    def __init__(self, attr, eps):
        super().__init__(attr, eps, 'FPR')


class PredPrevalenceConstraint(Constraint):
    def __init__(self, val):
        super().__init__()
        self.val = val

    def add(self, m_vars, dataset, binarized_df):
        n_plus = float((dataset.target == 1).sum())
        m_vars['model'] += n_plus - m_vars['lplus'] + m_vars['lminus'] <= self.var * len(dataset.target)

    def __str__(self):
        return f"Predicted Prevalence <= {self.val}"


import cplex
import logging
import numpy as np

logging.basicConfig(level='INFO')
log = logging.getLogger()

def get_sense(op):
    if op == '<=':
        sense = 'L'
    elif op == '==':
        sense = 'E'
    elif op == '>=':
        sense = 'G'
    else:
        raise NotImplementedError    
    return sense

def get_min_max(op, val, upper_limit):
    if val > upper_limit:
        val = upper_limit
    
    if op == '<=':
        return 1.0, float(val)
    elif op == '==':
        return float(val), float(val)
    elif op == '>=':
        return float(val), float(upper_limit)

class Constraint():
    def __init__(self):
        pass
    
    def add(self, *args, **kwargs):
        pass

class FNRConstraint(Constraint):
    def __init__(self, val):
        super().__init__()
        self.val = val
        
    def add(self, names, dataset, binarized_df, *args, **kwargs):
        return {
            'lin_expr': [cplex.SparsePair(ind = [names['lplus']], 
                                            val = [1.0])],
            'senses': ['L'],
            'rhs': [float((dataset.target == 1).sum() * self.val)]
        }
        
    def __str__(self):
        return f"FNR <= {self.val}"

class FPRConstraint(Constraint):
    def __init__(self, val):
        super().__init__()
        self.val = val
        
    def add(self, names, dataset, binarized_df, *args, **kwargs):
        return {
            'lin_expr': [cplex.SparsePair(ind = [names['lminus']], 
                                            val = [1.0])],
            'senses': ['L'],
            'rhs': [float((dataset.target == 0).sum() * self.val)]
        }
        
    def __str__(self):
        return f"FPR <= {self.val}"
    
class MaxNumFeatureConstraint(Constraint): # constraint on N
    def __init__(self, op, val):
        super().__init__()
        self.op = op
        self.val = val
    
    def add(self, names, *args, **kwargs):
        log.warn("Calling add on N_max constraint; should pass to Checklist_MIP.build_problem() instead")
        sense = get_sense(self.op) 
        return {
            'lin_expr': [cplex.SparsePair(ind = [names['N']], 
                                            val = [1.0])],
            'senses': [sense],
            'rhs': [float(self.val)]                                
        }           
        
    def get_min_max(self, n_features):
        return get_min_max(self.op, self.val, n_features)
        
    def __str__(self):
        return f"Total number of selected attributes {self.op} {self.val}."
    
class MConstraint(Constraint): # constraint on M
    def __init__(self, op, val):
        super().__init__()
        self.op = op
        self.val = val
    
    def add(self, names, *args, **kwargs):
        log.warn("Calling add on M_max constraint; should pass to Checklist_MIP.build_problem() instead")
        sense = get_sense(self.op)   
        return {
            'lin_expr':  [cplex.SparsePair(ind = [names['M']], 
                                            val = [1.0])],
            'senses': [sense],
            'rhs': [float(self.val)]                                
        }          
   
    def get_min_max(self, n_features):
        return get_min_max(self.op, self.val, n_features)
    
    def __str__(self):
        return f"M {self.op} {self.val}."        
    
class LossConstraint(Constraint): 
    def __init__(self, op = ">=", l = 1):
        self.l = l
        self.op = op
        
    def add(self, names, *args, **kwargs):
        sense = get_sense(self.op)  
        return {
            'lin_expr':   [cplex.SparsePair(ind = [names['l']], 
                                            val = [1.0])],
            'senses': [sense],
            'rhs': [float(self.l)]                                
        }   
            
    def __str__(self):
        return f"Total loss {self.op} {self.l}."
        
class FeaturesPerGroupConstraint(Constraint): # select N features from a particular feature group (ex: age)
    def __init__(self, feature, op = "<=", N = 1):
        self.N = N
        self.feature = feature
        self.op = op
        
    def add(self, names, dataset, binarized_df, *args, **kwargs):
        assert(self.feature in dataset.feature_mapping), f"{self.feature} not in original dataframe!"
        inds, accuracies = [], [] # we use the accuracy of a single feature as the weight for the SOS
        for c, i in enumerate(binarized_df.columns):
            if i in dataset.feature_mapping[self.feature].binarized_names:
                inds.append(c)
                accuracies.append( float((dataset.binarized_df[i] == dataset.target).sum()) )
        sense = get_sense(self.op)  
        if self.N > 1 or self.op != '<=':
            return {
                'lin_expr':[cplex.SparsePair(ind = [names['lamb'][i] for i in inds], 
                                                val = [1.0] * len(inds))],
                'senses': [sense],
                'rhs': [float(self.N)]
            }   
        else:
            return {
                'SOS': cplex.SparsePair(ind = [names['lamb'][i] for i in inds], 
                                                val = np.argsort(np.argsort(accuracies)).astype(float).tolist()),
                'name': f'one_from_{self.feature}'
            } 
             
                        
    def __str__(self):
        return f"Total number of attributes from group '{self.feature}' {self.op} {self.N}."
    
class MostOneofTwoConstraint(Constraint): # given names of two binarized features, select at most one of the two
    def __init__(self, n1, n2):
        super().__init__()
        self.n1 = n1
        self.n2 = n2
    
    def add(self, names, dataset, binarized_df, *args, **kwargs):
        try:
            ind1 = list(binarized_df.columns).index(self.n1) 
            ind2 = list(binarized_df.columns).index(self.n2) 
            # return {
            #     'lin_expr': [cplex.SparsePair(ind = [names['lamb'][ind1],names['lamb'][ind2]], 
            #                                 val = [1.0, 1.0])],
            #     'senses': ['L'],
            #     'rhs': [1.0]
            # }
            acc1 = float((dataset.binarized_df[self.n1] == dataset.target).sum())
            acc2 = float((dataset.binarized_df[self.n2] == dataset.target).sum())
            return {
                'SOS': cplex.SparsePair(ind = [names['lamb'][ind1],names['lamb'][ind2]], 
                                             val = [float(acc1 >= acc2), float(acc1 < acc2)]),
                'name': f"at_most_{self.n1}_or_{self.n2}"     
            }
            
        except ValueError: # 1 or more of the features were removed because of duplication
            return None
        
    def __str__(self):
        return f"One of {self.n1}, {self.n2}"      
    
class GroupMetricGapConstraintInterface(Constraint):
    def __init__(self, attr, eps, metric = 'FPR'):
        super().__init__()
        self.attr = attr
        self.eps = eps
        assert metric in ['FPR', 'FNR']
        self.metric = metric

    def add(self, names, dataset, binarized_df, *args, **kwargs):
        assert(self.attr in dataset.protected_attrs)
        grp_vals = dataset.protected_groups[self.attr].unique()
        linexprs = []
        if self.metric == "FPR":
            target_int = -1
        elif self.metric == 'FNR':
            target_int = 1        

        for c, g1 in enumerate(grp_vals):
            m1 = (dataset.protected_groups[self.attr] == g1).values
            n1 = (dataset.target.values == target_int)[m1].sum()
            inds1 = [i for c, i in enumerate(names['z']) if m1[c] and dataset.target.values[c] == target_int]

            for g2 in grp_vals[c+1:]:
                if g1 != g2:
                    m2 = (dataset.protected_groups[self.attr] == g2).values
                    n2 = (dataset.target.values == target_int)[m2].sum()
                    inds2 = [i for c, i in enumerate(names['z']) if m2[c] and dataset.target.values[c] == target_int]                

                    linexprs.append(
                            cplex.SparsePair(ind = inds1 + inds2, 
                                                    val = [1/n1]*len(inds1) + [-1/n2]*len(inds2))
                    )

                    linexprs.append(
                            cplex.SparsePair(ind = inds1 + inds2, 
                                                    val = [-1/n1]*len(inds1) + [1/n2]*len(inds2))
                    )
        
        return {
                'lin_expr': linexprs,
                'senses': ['L'] * len(linexprs),
                'rhs': [float(self.eps)] * len(linexprs)
            } 
    
    def __str__(self):
        return f"|{self.metric}(G1) - {self.metric}(G2)| <= {self.eps} for all (G1, G2) in {self.attr}"


class GroupFPRGapConstraint(GroupMetricGapConstraintInterface):
    def __init__(self, attr, eps):
        super().__init__(attr, eps, 'FPR')

class GroupFNRGapConstraint(GroupMetricGapConstraintInterface):
    def __init__(self, attr, eps):
        super().__init__(attr, eps, 'FNR')      

class GroupMetricConstraintInterface(Constraint):
    def __init__(self, attr, eps, metric = 'FNR'):
        super().__init__()
        self.attr = attr
        self.eps = eps
        assert metric in ['FPR', 'FNR']
        self.metric = metric

    def add(self, names, dataset, binarized_df, *args, **kwargs):
        assert(self.attr in dataset.protected_attrs)
        assert(len(names['z']) == len(dataset))
        grp_vals = dataset.protected_groups[self.attr].unique()
        linexprs = []
        if self.metric == "FPR":
            target_int = -1
        elif self.metric == 'FNR':
            target_int = 1  

        for g in grp_vals:
            m = (dataset.protected_groups[self.attr] == g).values            
            n = (dataset.target.values ==  target_int)[m].sum()
            inds = [i for c, i in enumerate(names['z']) if m[c] and dataset.target.values[c] ==  target_int]
            linexprs.append(
                    cplex.SparsePair(ind = inds, 
                                            val = [1/n]*len(inds))
            )
        return {
                'lin_expr': linexprs,
                'senses': ['L'] * len(linexprs),
                'rhs': [float(self.eps)] * len(linexprs)
            } 
    
    def __str__(self):
        return f"{self.metric}(G) <= {self.eps} for all (G) in {self.attr}"

class GroupFNRConstraint(Constraint):
    def __init__(self, attr, eps):
        super().__init__(attr, eps, 'FNR')
        
class GroupFPRConstraint(Constraint):
    def __init__(self, attr, eps):
        super().__init__(attr, eps, 'FPR')

class PredPrevalenceConstraint(Constraint):
    def __init__(self, val):
        super().__init__()
        self.val = val

    def add(self, names, dataset, binarized_df, *args, **kwargs):
        n_plus = float((dataset.target == 1).sum())
        return {
            'lin_expr': [cplex.SparsePair(ind = [names['lminus'], names['lplus']],
                                          val = [1.0, -1.0])],
            'senses': ['L'],
            'rhs': [float(len(dataset.target) * self.val - n_plus)]
        }

    def __str__(self):
        return f"Predicted Prevalence <= {self.val}"